# -*- coding: utf8 -*-
from __future__ import unicode_literals
import datetime
import logging

from driverng import MessageV2, Worklist
from driverng.splitted import Multiple, Splitted, String, Date

logger = logging.getLogger("driverng.lumipulse_g600ii")


class LumipulseBase(Splitted):
    SEPARATOR = "|"


class LumipulseOrder(LumipulseBase):
    FIELDS = (
        String("type", default="W"),  # W
        String("instrument"),
        String("carousel_no"),
        String("sample_id"),
        String("specimen"),
        String("manual_dilution"),
        Multiple(
            "tests",
            (String("test_code"), String("auto_dilution")),
            separator=LumipulseBase.SEPARATOR,
        ),
    )


class LumipulseQuery(LumipulseBase):
    FIELDS = [
        String("type"),  # R
        String("mode"),
        String("instrument"),
        String("carousel_no"),
        String("sample_id"),
    ]

    def parse(self, line):
        vals = line.split(self.SEPARATOR)
        for idx, f in enumerate(self.FIELDS):
            setattr(self, f.name, vals[idx])


class LumipulseResult(LumipulseBase):
    FIELDS = [
        String("type"),
        String("instrument"),
        String("sample_category"),
        String("batch_no"),
        String("sequence_no"),
        String("carousel_no"),
        String("sample_id"),
        String("specimen"),
        String("dilution"),
        String("nb_order_analyte"),
        String("order_analyte_no"),
        String("nb_multiple_assay"),
        String("multiple_assay_no"),
        String("analyte_no"),
        String("count_multiple_assay"),
        String("concentration_multiple_assay"),
        String("average_count_multiple"),
        String("calculated_concentration"),
        String("units"),
        String("error_analyte"),
        String("error_sample"),
        String("auto_dilution"),
        String("reagent_lot"),
        String("subtrate_lot"),
        Date("assay_date", fmt_date="%Y%m%d"),
        String("assay_time"),
        String("calibration_date"),
        String("calibration_time"),
    ]


class LumipulseMultiple(list):
    EXTENDED_RESULT = 99
    _MESSAGE_TYPE_MAP = {
        MessageV2.ORDER: LumipulseOrder,
        MessageV2.RESULT: LumipulseResult,
        MessageV2.QUERY: LumipulseQuery,
    }
    SEPARATOR = "|"

    def add(self, type_message=None, **kwargs):
        obj = self._MESSAGE_TYPE_MAP[type_message](**kwargs)
        self.append(obj)
        return obj

    def parse(self, data):
        for line in data.split("\n"):
            if not len(line):
                continue
            if line[-1] == self.SEPARATOR:
                line = line[:-1]
            champs = line.split(self.SEPARATOR)
            if champs[0] == "D":
                type_message = MessageV2.RESULT
            elif champs[0] == "R":
                type_message = MessageV2.QUERY
            elif champs[0] == "W":
                type_message = MessageV2.ORDER
            else:
                logger.warning("message non reconnu: %s", line)
                continue
            obj = self.add(type_message=type_message)
            obj.parse(line)

    def to_string(self):
        return "\r\n".join([e.to_string() for e in self])


class Driver(MessageV2):
    SPEC = {"code": "lumipulse_g600ii", "label": "Lumipulse G600II de Fujirebio"}

    MESSAGE_TYPE_MAP = {
        MessageV2.ORDER: LumipulseOrder,
        MessageV2.RESULT: LumipulseResult,
    }

    def create(self):
        return LumipulseMultiple()

    def to_string(self, message):
        return message.to_string()

    def parse(self, data):
        obj = LumipulseMultiple()
        obj.parse(data)
        return obj

    def encode_worklist(self, worklist, context):
        message = self.create()
        samples = worklist.get_worklist(
            unique_together=["action"],
            action={"del": worklist.IGNORE, worklist.ELSE: ""},
        )
        if len(samples) == 0:
            return None
        for sample, action in samples:
            order_message = message.add(
                MessageV2.ORDER, sample_id=sample.sample_id, specimen=sample.specimen
            )
            if context["query"]:
                order_message.instrument = context["query"].queries[0].intrument
                order_message.carousel_no = context["query"].queries[0].carousel_no
            for order in sample["orders"]:
                order_message.tests.add(
                    test_code=order["code"], auto_dilution=order.dilution or "1"
                )
        return message

    def decode_message(self, message):
        if message and isinstance(message[0], (LumipulseResult,)):
            # Format RESULT FILE
            worklist = Worklist(context="result")
            for m in message:
                if m.sample_category not in ("N", "Q"):
                    logger.warning("sample category non reconnu %s!", m.sample_category)
                    continue
                date_resu = None
                if m.assay_date:
                    if m.assay_time:
                        try:
                            time_resu = datetime.time(
                                int(m.assay_time[:2]),
                                int(m.assay_time[2:4]),
                                int(m.assay_time[4:]),
                            )
                            date_resu = m.assay_date.combine(m.assay_date, time_resu)
                        except (ValueError, IndexError):
                            date_resu = m.assay_date
                sample = worklist.get_or_create(
                    sample_id=m.sample_id,
                    specimen=m.specimen,
                    type="patient" if m.sample_category == "N" else "control",
                )
                sample.rack = "{}_{}".format(m.sequence_no, m.carousel_no)
                flags = []
                if m.error_analyte.strip():
                    flags.append(m.error_analyte.strip())
                if m.error_sample.strip() and m.error_sample.strip() not in flags:
                    flags.append(m.error_sample.strip())
                for test, value in [
                    ("COUNT", m.count_multiple_assay),
                    ("CONCENTRATION", m.concentration_multiple_assay),
                    ("AVERAGE", m.average_count_multiple),
                    ("CALCULATED", m.calculated_concentration),
                ]:
                    if value:
                        code = "{}{}".format(m.analyte_no, test)
                        resu = sample.results.add(
                            code=code,
                            value=value,
                            flags=flags,
                            dilution=m.auto_dilution,
                            unit=m.units.strip(),
                            instrument="g600ii" + m.instrument,
                            date=date_resu,
                        )
                        if m.reagent_lot:
                            resu.lots = [
                                {
                                    "type": "reagent",
                                    "lot": m.reagent_lot + "_" + m.subtrate_lot,
                                }
                            ]

        elif message and isinstance(message[0], LumipulseQuery):
            worklist = Worklist(context="query")
            for m in message:
                if m.mode == "S":
                    worklist.get_or_create(type="patient", sample_id=m.sample_id)
        return worklist
